@@ -12,6 +12,7 @@ from typing import (
     TypeAlias,
     Union,
     cast,
+    overload,
 )
 
 import aiofiles
@@ -32,18 +33,25 @@ from pydoll.exceptions import (
     IFrameNotFound,
     InvalidFileExtension,
     InvalidIFrame,
+    InvalidScriptWithElement,
+    NetworkEventsNotEnabled,
     NoDialogPresent,
     NotAnIFrame,
     PageLoadTimeout,
     WaitElementTimeout,
 )
 from pydoll.protocol.base import Response
-from pydoll.protocol.network.types import Cookie, CookieParam
+from pydoll.protocol.network.responses import GetResponseBodyResponse
+from pydoll.protocol.network.types import Cookie, CookieParam, NetworkLog
 from pydoll.protocol.page.events import PageEvent
 from pydoll.protocol.page.responses import CaptureScreenshotResponse, PrintToPDFResponse
-from pydoll.protocol.runtime.responses import EvaluateResponse
+from pydoll.protocol.runtime.responses import CallFunctionOnResponse, EvaluateResponse
 from pydoll.protocol.storage.responses import GetCookiesResponse
-from pydoll.utils import decode_base64_to_bytes
+from pydoll.utils import (
+    decode_base64_to_bytes,
+    has_return_outside_function,
+    is_script_already_function,
+)
 
 if TYPE_CHECKING:
     from pydoll.browser.chromium.base import Browser
@@ -60,8 +68,43 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
     Primary interface for web page automation including navigation, DOM manipulation,
     JavaScript execution, event handling, network monitoring, and specialized tasks
     like Cloudflare bypass.
+
+    This class implements a singleton pattern based on target_id to ensure
+    only one Tab instance exists per browser tab.
     """
 
+    _instances: dict[str, 'Tab'] = {}
+
+    def __new__(
+        cls,
+        browser: 'Browser',
+        connection_port: int,
+        target_id: str,
+        browser_context_id: Optional[str] = None,
+    ) -> 'Tab':
+        """
+        Create or return existing Tab instance for the given target_id.
+
+        Args:
+            browser: Browser instance that created this tab.
+            connection_port: CDP WebSocket port.
+            target_id: CDP target identifier for this tab.
+            browser_context_id: Optional browser context ID.
+
+        Returns:
+            Tab instance (new or existing) for the target_id.
+        """
+        if target_id in cls._instances:
+            existing_instance = cls._instances[target_id]
+            existing_instance._browser = browser
+            existing_instance._connection_port = connection_port
+            existing_instance._browser_context_id = browser_context_id
+            return existing_instance
+
+        instance = super().__new__(cls)
+        cls._instances[target_id] = instance
+        return instance
+
     def __init__(
         self,
         browser: 'Browser',
@@ -78,18 +121,57 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             target_id: CDP target identifier for this tab.
             browser_context_id: Optional browser context ID.
         """
-        self._browser = browser
-        self._connection_port = connection_port
-        self._target_id = target_id
-        self._connection_handler = ConnectionHandler(connection_port, self._target_id)
-        self._page_events_enabled = False
-        self._network_events_enabled = False
-        self._fetch_events_enabled = False
-        self._dom_events_enabled = False
-        self._runtime_events_enabled = False
-        self._intercept_file_chooser_dialog_enabled = False
+        if hasattr(self, '_initialized') and self._initialized:
+            return
+
+        self._browser: 'Browser' = browser
+        self._connection_port: int = connection_port
+        self._target_id: str = target_id
+        self._connection_handler: ConnectionHandler = ConnectionHandler(
+            connection_port, self._target_id
+        )
+        self._page_events_enabled: bool = False
+        self._network_events_enabled: bool = False
+        self._fetch_events_enabled: bool = False
+        self._dom_events_enabled: bool = False
+        self._runtime_events_enabled: bool = False
+        self._intercept_file_chooser_dialog_enabled: bool = False
         self._cloudflare_captcha_callback_id: Optional[int] = None
-        self._browser_context_id = browser_context_id
+        self._browser_context_id: Optional[str] = browser_context_id
+        self._initialized: bool = True
+
+    @classmethod
+    def _remove_instance(cls, target_id: str) -> None:
+        """
+        Remove instance from registry when tab is closed.
+
+        Args:
+            target_id: Target ID to remove from registry.
+        """
+        cls._instances.pop(target_id, None)
+
+    @classmethod
+    def get_instance(cls, target_id: str) -> Optional['Tab']:
+        """
+        Get existing Tab instance for target_id if it exists.
+
+        Args:
+            target_id: Target ID to look up.
+
+        Returns:
+            Existing Tab instance or None if not found.
+        """
+        return cls._instances.get(target_id)
+
+    @classmethod
+    def get_all_instances(cls) -> dict[str, 'Tab']:
+        """
+        Get all active Tab instances.
+
+        Returns:
+            Dictionary mapping target_id to Tab instances.
+        """
+        return cls._instances.copy()
 
     @property
     def page_events_enabled(self) -> bool:
@@ -212,9 +294,6 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             custom_selector: Custom captcha selector (default: cf-turnstile class).
             time_before_click: Delay before clicking captcha (default 2s).
             time_to_wait_captcha: Timeout for captcha detection (default 5s).
-
-        Returns:
-            Callback ID for disabling auto-solver.
         """
         if not self.page_events_enabled:
             await self.enable_page_events()
@@ -278,14 +357,16 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
         Note:
             Tab instance becomes invalid after calling this method.
         """
-        return await self._execute_command(PageCommands.close())
+        result = await self._execute_command(PageCommands.close())
+        self._remove_instance(self._target_id)
+        return result
 
     async def get_frame(self, frame: WebElement) -> IFrame:
         """
         Get Tab object for interacting with iframe content.
 
         Args:
-            frame: Tab representing the iframe (<iframe> tag).
+            frame: Tab representing the iframe tag.
 
         Returns:
             Tab instance configured for iframe interaction.
@@ -295,7 +376,6 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             InvalidIFrame: If iframe lacks valid src attribute.
             IFrameNotFound: If iframe target not found in browser.
         """
-        print('frame.tag_name: ', frame.tag_name)
         if not frame.tag_name == 'iframe':
             raise NotAnIFrame
 
@@ -317,6 +397,50 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
         )
         return response['result']['cookies']
 
+    async def get_network_response_body(self, request_id: str) -> str:
+        """
+        Get the response body for a given request ID.
+
+        Args:
+            request_id: Request ID to get the response body for.
+
+        Returns:
+            The response body for the given request ID.
+
+        Raises:
+            NetworkEventsNotEnabled: If network events are not enabled.
+        """
+        if not self.network_events_enabled:
+            raise NetworkEventsNotEnabled('Network events must be enabled to get response body')
+
+        response: GetResponseBodyResponse = await self._execute_command(
+            NetworkCommands.get_response_body(request_id)
+        )
+        return response['result']['body']
+
+    async def get_network_logs(self, filter: Optional[str] = None) -> list[NetworkLog]:
+        """
+        Get network logs.
+
+        Args:
+            filter: Filter to apply to the network logs.
+
+        Returns:
+            The network logs.
+
+        Raises:
+            NetworkEventsNotEnabled: If network events are not enabled.
+        """
+        if not self.network_events_enabled:
+            raise NetworkEventsNotEnabled('Network events must be enabled to get network logs')
+
+        logs = self._connection_handler.network_logs
+        if filter:
+            logs = [
+                log for log in logs if filter in log['params'].get('request', {}).get('url', '')
+            ]
+        return logs
+
     async def set_cookies(self, cookies: list[CookieParam]):
         """
         Set multiple cookies for current page.
@@ -512,7 +636,15 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             PageCommands.handle_javascript_dialog(accept=accept, prompt_text=prompt_text)
         )
 
-    async def execute_script(self, script: str, element: Optional[WebElement] = None):
+    @overload
+    async def execute_script(self, script: str) -> EvaluateResponse: ...
+
+    @overload
+    async def execute_script(self, script: str, element: WebElement) -> CallFunctionOnResponse: ...
+
+    async def execute_script(
+        self, script: str, element: Optional[WebElement] = None
+    ) -> Union[EvaluateResponse, CallFunctionOnResponse]:
         """
         Execute JavaScript in page context.
 
@@ -523,17 +655,17 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
         Examples:
             await page.execute_script('argument.click()', element)
             await page.execute_script('argument.value = "Hello"', element)
+
+        Raises:
+            InvalidScriptWithElement: If script contains 'argument' but no element is provided.
         """
+        if 'argument' in script and element is None:
+            raise InvalidScriptWithElement('Script contains "argument" but no element was provided')
+
         if element:
-            script = script.replace('argument', 'this')
-            script = f'function(){{ {script} }}'
-            object_id = element._object_id
-            command = RuntimeCommands.call_function_on(
-                object_id=object_id, function_declaration=script, return_by_value=True
-            )
-        else:
-            command = RuntimeCommands.evaluate(expression=script)
-        return await self._execute_command(command)
+            return await self._execute_script_with_element(script, element)
+
+        return await self._execute_script_without_element(script)
 
     @asynccontextmanager
     async def expect_file_chooser(
@@ -616,7 +748,12 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             if not _before_page_events_enabled:
                 await self.disable_page_events()
 
-    async def on(self, event_name: str, callback: Callable[[dict], Any], temporary: bool = False):
+    async def on(
+        self,
+        event_name: str,
+        callback: Callable[[dict], Any],
+        temporary: bool = False,
+    ) -> int:
         """
         Register CDP event listener.
 
@@ -646,6 +783,46 @@ class Tab(FindElementsMixin):  # noqa: PLR0904
             event_name, function_to_register, temporary
         )
 
+    async def _execute_script_with_element(self, script: str, element: WebElement):
+        """
+        Execute script with element context.
+
+        Args:
+            script: JavaScript code to execute.
+            element: Element context (use 'argument' in script to reference).
+
+        Returns:
+            The result of the script execution.
+        """
+        if 'argument' not in script:
+            raise InvalidScriptWithElement('Script does not contain "argument"')
+
+        script = script.replace('argument', 'this')
+
+        if not is_script_already_function(script):
+            script = f'function(){{ {script} }}'
+
+        command = RuntimeCommands.call_function_on(
+            object_id=element._object_id, function_declaration=script, return_by_value=True
+        )
+        return await self._execute_command(command)
+
+    async def _execute_script_without_element(self, script: str):
+        """
+        Execute script without element context.
+
+        Args:
+            script: JavaScript code to execute.
+
+        Returns:
+            The result of the script execution.
+        """
+        if has_return_outside_function(script):
+            script = f'(function(){{ {script} }})()'
+
+        command = RuntimeCommands.evaluate(expression=script)
+        return await self._execute_command(command)
+
     async def _refresh_if_url_not_changed(self, url: str) -> bool:
         """Refresh page if URL hasn't changed."""
         current_url = await self.current_url
